import re
from typing import List, Union

camel_to_snake_regex = re.compile(r'(?<!^)(?=[A-Z])')


def camel_to_snake(word: str) -> str:
    return camel_to_snake_regex.sub('_', word).lower()


def find_section_entries(lines: List[str]) -> List[int]:
    """Find the strings which contains a section entry header."""
    section_idx = [
        idx
        for idx, line
        in enumerate(lines)
        if (
            line.startswith('[ZValue =')
            or line.startswith('[MontSection =')
            or line.startswith('[FrameSet =')
        )
    ]

    return section_idx


def find_title_entries(lines: List[str]) -> List[int]:
    """Find mdoc title entries in a list of strings"""
    title_idxs = []
    for idx, line in enumerate(lines):
        if line.startswith('[T ='):
            title_idxs.append(idx)
    return title_idxs


from pathlib import Path

import pytest


@pytest.fixture
def tilt_series_mdoc_file():
    return Path(__file__).parent / 'test_data' / 'tilt_series.mdoc'

@pytest.fixture
def tilt_series_mdoc_string():
    with open(Path(__file__).parent / 'test_data' / 'tilt_series.mdoc') as f:
        return f.read()

@pytest.fixture
def montage_section_mdoc_file():
    return Path(__file__).parent / 'test_data' / 'montage_section.mdoc'


@pytest.fixture
def montage_section_multiple_mdoc_file():
    return Path(__file__).parent / 'test_data' / 'montage_section_multiple.mdoc'


@pytest.fixture
def frame_set_single_mdoc_file():
    return Path(__file__).parent / 'test_data' / 'frame_set_single.mdoc'


@pytest.fixture
def frame_set_multiple_mdoc_file():
    return Path(__file__).parent / 'test_data' / 'frame_set_multiple.mdoc'



import pandas as pd

from mdocfile import read
from mdocfile.data_models import Mdoc


def test_read_tilt_series_mdoc(tilt_series_mdoc_file):
    df = read(tilt_series_mdoc_file)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (41, 26)
    assert 'TiltAngle' in df.columns

def test_read_tilt_series_mdoc_string(tilt_series_mdoc_string):
    df = Mdoc.from_string(tilt_series_mdoc_string).as_dataframe()
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (41, 26)
    assert 'TiltAngle' in df.columns

def test_read_montage_section_mdoc(montage_section_mdoc_file):
    df = read(montage_section_mdoc_file)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (63, 37)


def test_read_montage_section_multiple_mdoc(montage_section_multiple_mdoc_file):
    df = read(montage_section_multiple_mdoc_file)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (100, 36)


def test_read_frame_set_single_mdoc(frame_set_single_mdoc_file):
    df = read(frame_set_single_mdoc_file)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (1, 26)


def test_read_frame_set_multiple_mdoc(frame_set_multiple_mdoc_file):
    df = read(frame_set_multiple_mdoc_file)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (21, 28)



from os import PathLike

import pandas as pd

from .data_models import Mdoc


def read(filename: PathLike) -> pd.DataFrame:
    """Read an mdoc file as a pandas dataframe.

    Parameters
    ----------
    filename : PathLike
        SerialEM mdoc file to read

    Returns
    -------
    df : pd.DataFrame
        dataframe containing info from mdoc file
    """
    return Mdoc.from_file(filename).as_dataframe()



from pathlib import Path, PureWindowsPath
from tempfile import NamedTemporaryFile

from mdocfile.data_models import MdocGlobalData, MdocSectionData, Mdoc

GLOBAL_DATA_EXAMPLE = r"""PixelSpacing = 5.4
ImageFile = TS_01.mrc
ImageSize = 924 958
DataMode = 1
"""


def test_global_data_from_lines():
    lines = GLOBAL_DATA_EXAMPLE.split('\n')
    data = MdocGlobalData.from_lines(lines)
    assert isinstance(data, MdocGlobalData)
    assert data.PixelSpacing == 5.4
    assert data.ImageFile == Path('TS_01.mrc')
    assert data.DataMode == 1


SECTION_DATA_EXAMPLE = r"""[ZValue = 0]
TiltAngle = 0.000999877
StagePosition = 20.7936 155.287
StageZ = 163.803
Magnification = 105000
Intensity = 0.00900259
ExposureDose = 0
PixelSpacing = 5.4
SpotSize = 8
Defocus = 2.68083
ImageShift = -0.0108126 -0.121079
RotationAngle = 175.3
ExposureTime = 0.8
Binning = 0.5
CameraIndex = 2
DividedBy2 = 1
MagIndex = 31
MinMaxMean = 5 1403 623.699
TargetDefocus = -4
SubFramePath = D:\DATA\Flo\HGK149_20151130\frames\TS_01_000_0.0.mrc
NumSubFrames = 8
DateTime = 30-Nov-15  15:21:38
"""


def test_section_data_from_lines():
    lines = SECTION_DATA_EXAMPLE.split('\n')
    data = MdocSectionData.from_lines(lines)
    assert isinstance(data, MdocSectionData)
    assert data.TiltAngle == 0.000999877
    assert data.StagePosition == (20.7936, 155.287)
    assert data.StageZ == 163.803
    assert data.Magnification == 105000
    assert data.Intensity == 0.009002590
    assert data.ExposureDose == 0
    assert data.PixelSpacing == 5.4
    assert data.SpotSize == 8
    assert data.Defocus == 2.68083
    assert data.ImageShift == (-0.0108126, -0.121079)
    assert data.RotationAngle == 175.3
    assert data.ExposureTime == 0.8
    assert data.Binning == 0.5
    assert data.CameraIndex == 2
    assert data.DividedBy2 is True
    assert data.MagIndex == 31
    assert data.MinMaxMean == (5, 1403, 623.699)
    assert data.TargetDefocus == -4
    assert data.SubFramePath == PureWindowsPath(r'D:\DATA\Flo\HGK149_20151130\frames\TS_01_000_0.0.mrc')
    assert data.NumSubFrames == 8
    assert data.DateTime == '30-Nov-15  15:21:38'


def test_mdoc_from_tilt_series_mdoc_file(tilt_series_mdoc_file):
    mdoc = Mdoc.from_file(tilt_series_mdoc_file)
    assert isinstance(mdoc, Mdoc)
    assert len(mdoc.titles) == 2
    assert mdoc.global_data.PixelSpacing == 5.4
    assert len(mdoc.section_data) == 41


def test_to_string_is_valid_mdoc(tilt_series_mdoc_file):
    mdoc = Mdoc.from_file(tilt_series_mdoc_file)
    with NamedTemporaryFile() as tmp:
        tmp.write(mdoc.to_string().encode())
        mdoc2 = Mdoc.from_file(tmp.name)
    mdoc_dict = mdoc.section_data[0].model_dump()
    mdoc2_dict = mdoc2.section_data[0].model_dump()
    for (k1, v1), (k2, v2) in zip(mdoc_dict.items(), mdoc2_dict.items()):
        assert v1 == v2
        assert k1 == k2

def test_section_data_from_path():
    section = MdocSectionData(SubFramePath=Path('bla.tif'))


import pandas as pd
from pydantic import field_validator, BaseModel
from pathlib import Path, PureWindowsPath
from typing import List, Optional, Tuple, Union, Sequence

from mdocfile.utils import find_section_entries, find_title_entries


class MdocGlobalData(BaseModel):
    """Data model for global data in a SerialEM mdoc file.

    https://bio3d.colorado.edu/SerialEM/hlp/html/about_formats.htm
    """
    DataMode: Optional[int] = None
    ImageSize: Optional[Tuple[int, int]] = None
    Montage: Optional[bool] = None
    ImageSeries: Optional[int] = None
    ImageFile: Optional[Path] = None
    PixelSpacing: Optional[float] = None
    Voltage: Optional[float] = None

    @field_validator('ImageSize', mode="before")
    @classmethod
    def multi_number_string_to_tuple(cls, value: str):
        if isinstance(value, str):
            value = tuple(value.split())
        return value

    @classmethod
    def from_lines(cls, lines: List[str]):
        lines = [
            line for line in lines
            if len(line) > 0
        ]
        key_value_pairs = [
            line.split('=') for line in lines
            if not line.startswith('[T =')
        ]
        key_value_pairs = [
            (k.strip(), v.strip()) for k, v in key_value_pairs
        ]
        data = {k: v for k, v in key_value_pairs}
        return cls(**data)

    def to_string(self):
        lines = []
        for k, v in self.model_dump().items():
            if v is None:
                continue
            if isinstance(v, tuple):
                v = ' '.join(str(el) for el in v)
            if v == 'nan':
                v = 'NaN'
            lines.append(f'{k} = {v}')
        return '\n'.join(lines)


class MdocSectionData(BaseModel):
    """Data model for section data in a SerialEM mdoc file.

    https://bio3d.colorado.edu/SerialEM/hlp/html/about_formats.htm
    """
    # headers
    ZValue: Optional[int] = None
    MontSection: Optional[int] = None
    FrameSet: Optional[int] = None

    # section data
    TiltAngle: Optional[float] = None
    PieceCoordinates: Optional[Tuple[float, float, int]] = None
    StagePosition: Optional[Tuple[float, float]] = None
    StageZ: Optional[float] = None
    Magnification: Optional[float] = None
    CameraLength: Optional[float] = None
    MagIndex: Optional[int] = None
    Intensity: Optional[float] = None
    SuperMontCoords: Optional[Tuple[float, float]] = None
    PixelSpacing: Optional[float] = None
    ExposureDose: Optional[float] = None
    DoseRate: Optional[float] = None
    SpotSize: Optional[float] = None
    Defocus: Optional[float] = None
    TargetDefocus: Optional[float] = None
    ImageShift: Optional[Tuple[float, float]] = None
    RotationAngle: Optional[float] = None
    ExposureTime: Optional[float] = None
    Binning: Optional[float] = None
    UsingCDS: Optional[bool] = None
    CameraIndex: Optional[int] = None
    DividedBy2: Optional[bool] = None
    LowDoseConSet: Optional[int] = None
    MinMaxMean: Optional[Tuple[float, float, float]] = None
    PriorRecordDose: Optional[float] = None
    XedgeDxy: Optional[Tuple[float, float]] = None
    YedgeDxy: Optional[Tuple[float, float]] = None
    XedgeDxyVS: Optional[Union[Tuple[float, float], Tuple[float, float, float]]] = None
    YedgeDxyVS: Optional[Union[Tuple[float, float], Tuple[float, float, float]]] = None
    StageOffsets: Optional[Tuple[float, float]] = None
    AlignedPieceCoords: Optional[Union[Tuple[float, float], Tuple[float, float, float]]] = None
    AlignedPieceCoordsVS: Optional[
        Union[Tuple[float, float], Tuple[float, float, float]]] = None
    SubFramePath: Optional[Union[PureWindowsPath, Path]] = None
    NumSubFrames: Optional[int] = None
    FrameDosesAndNumbers: Optional[Sequence[Tuple[float, int]]] = None
    DateTime: Optional[str] = None
    NavigatorLabel: Optional[str] = None
    FilterSlitAndLoss: Optional[Tuple[float, float]] = None
    ChannelName: Optional[str] = None
    MultiShotHoleAndPosition: Optional[Union[Tuple[int, int], Tuple[int, int, int]]] = None
    CameraPixelSize: Optional[float] = None
    Voltage: Optional[float] = None

    @field_validator(
        'PieceCoordinates',
        'SuperMontCoords',
        'ImageShift',
        'MinMaxMean',
        'StagePosition',
        'XedgeDxy',
        'YedgeDxy',
        'XedgeDxyVS',
        'YedgeDxyVS',
        'StageOffsets',
        'AlignedPieceCoords',
        'AlignedPieceCoordsVS',
        'FrameDosesAndNumbers',
        'FilterSlitAndLoss',
        'MultiShotHoleAndPosition',
        mode="before")
    @classmethod
    def multi_number_string_to_tuple(cls, value: str):
        if isinstance(value, str):
            value = tuple(value.split())
        return value

    @classmethod
    def from_lines(cls, lines: List[str]):
        lines = [line.strip('[]')
                 for line
                 in lines
                 if len(line) > 0]
        key_value_pairs = [line.split('=') for line in lines]
        key_value_pairs = [
            (k.strip(), v.strip())
            for k, v
            in key_value_pairs
        ]
        lines = {k: v for k, v in key_value_pairs}
        return cls(**lines)

    def to_string(self):
        data = self.model_dump()
        z_value = data.pop('ZValue')
        lines = [f'[ZValue = {z_value}]']
        for k, v in data.items():
            if v is None:
                continue
            elif isinstance(v, tuple):
                v = ' '.join(str(el) for el in v)
            elif v == 'nan':
                v = 'NaN'
            lines.append(f'{k} = {v}')
        return '\n'.join(lines)


class Mdoc(BaseModel):
    titles: List[str]
    global_data: MdocGlobalData
    section_data: List[MdocSectionData]

    @classmethod
    def from_file(cls, filename: str):
        with open(filename) as file:
            return cls.from_lines(file.readlines())
    
    @classmethod
    def from_string(cls, string: str):
        lines = string.split('\n')

        return cls.from_lines(lines)
    
    @classmethod
    def from_lines(cls, file_lines: List[str]) -> 'Mdoc':
        lines = [line.strip() for line in file_lines]
        split_idxs = find_section_entries(lines)
        split_idxs.append(len(lines))

        header_lines = lines[0:split_idxs[0]]
        title_idxs = find_title_entries(header_lines)

        titles = [header_lines[idx] for idx in title_idxs]
        global_data = MdocGlobalData.from_lines(header_lines)
        section_data = [
            MdocSectionData.from_lines(lines[start_idx:end_idx])
            for start_idx, end_idx
            in zip(split_idxs, split_idxs[1:])
        ]
        return cls(titles=titles, global_data=global_data, section_data=section_data)
    
    def as_dataframe(self) -> pd.DataFrame:
        """
        Convert an Mdoc object to a pandas DataFrame
        """
        global_data = self.global_data.model_dump()
        section_data = {
            k: [section.model_dump()[k] for section in self.section_data]
            for k
            in self.section_data[0].model_dump().keys()
        }
        df = pd.DataFrame(data=section_data)

        # add duplicate copies of global data and mdoc file titles to each row of
        # the dataframe - tidy data is easier to analyse
        for k, v in global_data.items():
            df[k] = [v] * len(df)
        df['titles'] = [self.titles] * len(df)
        df = df.dropna(axis='columns', how='all')
        return df

    def to_string(self):
        """
        Generate the string representation of the Mdoc data
        """
        return '\n\n'.join([
            self.global_data.to_string(),
            '\n\n'.join(self.titles),
            '\n\n'.join(section.to_string() for section in self.section_data),
        ])
